import logging
from collections.abc import Iterator, Mapping
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd

if TYPE_CHECKING:
    from pathlib import Path

LOGGER = logging.getLogger(__name__)

Run = Mapping[str, Mapping[str, float]]


def _attach_queries(df: pd.DataFrame, queries: Mapping[str, str]) -> pd.DataFrame:
    """Attach queries to a data frame.

    :param df: The data frame (same format as used by rankings).
    :param queries: Query IDs mapped to queries.
    :raises ValueError: When the queries are incomplete.
    :return: The data frame with queries attached.
    """
    if not set(pd.unique(df["q_id"])).issubset(set(queries.keys())):
        raise ValueError("Queries are incomplete.")
    return df.merge(
        pd.DataFrame(queries.items(), columns=["q_id", "query"]), how="left", on="q_id"
    )


def _add_ranks(df: pd.DataFrame) -> pd.DataFrame:
    """Add a new column ("rank") to a data frame.

    The new column contains ranks of documents based on the scores w.r.t. the queries.
    Ranks start from `1`.

    :param df: The data frame to add the column to.
    :return: The data frame with the new column added.
    """
    df_ranks = df.groupby("q_id").cumcount().to_frame() + 1
    df_ranks.columns = ("rank",)
    return df.join(df_ranks)


def _minmax_normalize_scores(df: pd.DataFrame) -> pd.DataFrame:
    """Return a copy of a data frame with normalized scores (min-max).

    If all scores are equal, they are set to `0`.

    :param df: The input data frame.
    :return: A copy of the data frame with normalized scores.
    """
    new_df = df.copy()
    min_val = new_df["score"].min()
    max_val = new_df["score"].max()
    if min_val == max_val:
        LOGGER.warning("all scores are equal, setting scores to 0")
        new_df["score"] = 0
    else:
        new_df["score"] = (new_df["score"] - min_val) / (max_val - min_val)
    return new_df


class Ranking:
    """Represents rankings of documents/passages w.r.t. queries."""

    def __init__(
        self,
        df: pd.DataFrame,
        name: str | None = None,
        queries: Mapping[str, str] | None = None,
        dtype: np.dtype = np.dtype(np.float32),
        copy: bool = True,
        is_sorted: bool = False,
    ) -> None:
        """Create a ranking from an existing data frame.

        The data frame must have the columns "q_id", "id", "score",
        and (optionally) "query".

        Rows with NaN scores are removed.

        :param df: Data frame containing IDs and scores.
        :param name: Method name.
        :param queries: Query IDs mapped to queries.
        :param dtype: How the scores should be represented in the data frame.
        :param copy: Whether to copy the data frame.
        :param is_sorted: Whether the data frame is already sorted (by score).
        :raises ValueError: When a query-document/passage pair appears more than once.
        :raises ValueError: When the queries are incomplete.
        """
        super().__init__()
        self.name = name

        if df.duplicated(subset=["q_id", "id"]).any():
            raise ValueError(
                "Only one score per query-document/passage pair is allowed."
            )

        cols = ["q_id", "id", "score"]
        if "query" in df.columns:
            cols += ["query"]
        self._df = df.loc[:, cols].dropna()
        if copy:
            self._df = self._df.copy()

        for col, dt in (
            ("score", dtype),
            ("q_id", str),
            ("id", str),
        ):
            if col in self._df.columns and self._df[col].dtype != dt:
                self._df[col] = self._df[col].astype(dt)

        if not is_sorted:
            self._df.sort_values(by=["q_id", "score"], inplace=True, ascending=False)
        self._df.reset_index(inplace=True, drop=True)

        self._q_ids = set(pd.unique(self._df["q_id"]))
        if queries is not None:
            self._df = _attach_queries(self._df, queries)

    @property
    def has_queries(self) -> bool:
        """Whether the ranking has queries attached.

        :return: Whether queries exist.
        """
        return "query" in self._df.columns

    @property
    def q_ids(self) -> set[str]:
        """The set of (unique) query IDs in this ranking.

        Only queries with at least one scored document are considered.

        :return: The query IDs.
        """
        return self._q_ids

    def __getitem__(self, q_id: str) -> dict[str, float]:
        """Return the ranking for a query.

        :param q_id: The query ID.
        :return: Document/passage IDs mapped to scores.
        """
        return dict(self._df[self._df["q_id"] == q_id][["id", "score"]].values)

    def __len__(self) -> int:
        """Return the number of queries.

        :return: The number of queries.
        """
        return len(self._q_ids)

    def __iter__(self) -> Iterator[str]:
        """Yield all query IDs.

        :return: Iterator yielding query IDs.
        """
        yield from self._q_ids

    def __contains__(self, key: object) -> bool:
        """Check whether a query ID has associated document/passage IDs.

        :param key: The query ID.
        :return: Whether the query ID is in the ranking.
        """
        return key in self._q_ids

    def __eq__(self, o: object) -> bool:
        """Check if this ranking is identical to another one.

        Only takes IDs and scores into account.

        :param o: The other ranking.
        :return: Whether the two rankings are identical.
        """
        if not isinstance(o, Ranking):
            return False

        df1 = self._df.sort_values(["q_id", "id"]).reset_index(drop=True)
        df2 = o._df.sort_values(["q_id", "id"]).reset_index(drop=True)

        cols = ["q_id", "id", "score"]
        return df1[cols].equals(df2[cols])

    def __add__(self, o: "Ranking | float") -> "Ranking":
        """Add a constant or the scores of another ranking to this ranking's scores.

        Missing scores in either ranking are treated as zero.

        :param o: A ranking or a constant.
        :return: The resulting ranking with added scores.
        """
        if isinstance(o, Ranking):
            new_df = self._df.merge(
                o._df, on=["q_id", "id"], suffixes=(None, "_other"), how="outer"
            ).fillna(0)
            new_df["score"] = new_df["score"] + new_df["score_other"]
            is_sorted = False
        elif isinstance(o, int | float):
            new_df = self._df.copy()
            new_df["score"] += o
            is_sorted = True
        else:
            return NotImplemented

        return Ranking(
            new_df,
            name=self.name,
            dtype=self._df.dtypes["score"],
            copy=False,
            is_sorted=is_sorted,
        )

    __radd__ = __add__

    def __mul__(self, o: float) -> "Ranking":
        """Multiply this ranking's scores by a constant.

        :param o: A constant.
        :return: The resulting ranking with multiplied scores.
        """
        if not isinstance(o, int | float):
            return NotImplemented

        new_df = self._df.copy()
        new_df["score"] *= o

        return Ranking(
            new_df,
            name=self.name,
            dtype=self._df.dtypes["score"],
            copy=False,
            is_sorted=True,
        )

    __rmul__ = __mul__

    def __repr__(self) -> str:
        """Return a string representation of this ranking.

        :return: The string representation.
        """
        return self._df.__repr__()

    def attach_queries(self, queries: Mapping[str, str]) -> "Ranking":
        """Attach queries to the ranking.

        :param queries: Query IDs mapped to queries.
        :raises ValueError: When the queries are incomplete.
        :return: The ranking with queries attached.
        """
        return Ranking(
            self._df,
            self.name,
            queries=queries,
            dtype=self._df.dtypes["score"],
            copy=True,
            is_sorted=True,
        )

    def normalize(self) -> "Ranking":
        """Normalize the scores (min-max) to be in `[0, 1]`.

        If all scores are equal, they are set to `0`.

        :return: A ranking with normalized scores.
        """
        return Ranking(
            _minmax_normalize_scores(self._df),
            self.name,
            dtype=self._df.dtypes["score"],
            copy=False,
            is_sorted=True,
        )

    def cut(self, cutoff: int) -> "Ranking":
        """For each query, remove all but the top-`k` scoring documents/passages.

        :param cutoff: Number of best scores per query to keep (`k`).
        :return: The resulting ranking.
        """
        return Ranking(
            self._df.groupby("q_id").head(cutoff).reset_index(drop=True),
            name=self.name,
            dtype=self._df.dtypes["score"],
            copy=True,
            is_sorted=True,
        )

    def interpolate(
        self,
        other: "Ranking",
        alpha: float,
        normalize: bool = False,
    ) -> "Ranking":
        """Interpolate as `score = self.score * alpha + other.score * (1 - alpha)`.

        Missing scores in either ranking are treated as zero.

        :param other: Ranking to interpolate with.
        :param alpha: Interpolation parameter.
        :param normalize: Perform min-max normalization.
        :return: The resulting ranking.
        """
        df1 = self._df if not normalize else _minmax_normalize_scores(self._df)
        df2 = other._df if not normalize else _minmax_normalize_scores(other._df)

        # during normalization the data frames are copied already
        new_df = df1.merge(
            df2,
            on=["q_id", "id"],
            suffixes=(None, "_other"),
            copy=not normalize,
            how="outer",
        ).fillna(0)
        new_df["score"] = alpha * new_df["score"] + (1 - alpha) * new_df["score_other"]
        return Ranking(
            new_df,
            name=self.name,
            dtype=self._df.dtypes["score"],
            copy=False,
            is_sorted=False,
        )

    def rr_scores(self, k: int = 60) -> "Ranking":
        """Re-score documents/passages using reciprocal rank.

        A score is computed as `1 / (rank + k)`.

        Used by [RRF](https://dl.acm.org/doi/10.1145/1571941.1572114).

        :param k: RR scoring parameter.
        :return: A ranking with RR scores.
        """
        new_df = _add_ranks(self._df)
        new_df["score"] = 1 / (new_df["rank"] + k)
        return Ranking(
            new_df,
            name=self.name,
            dtype=self._df.dtypes["score"],
            copy=False,
            is_sorted=True,
        )

    def save(
        self,
        target: "Path",
    ) -> None:
        """Save the ranking in a TREC runfile.

        :param target: Output file.
        """
        df_out = _add_ranks(self._df)
        df_out["name"] = str(self.name)
        df_out["q0"] = "Q0"
        target.parent.mkdir(parents=True, exist_ok=True)
        df_out.to_csv(
            target,
            sep="\t",
            columns=["q_id", "q0", "id", "rank", "score", "name"],
            index=False,
            header=False,
        )

    @classmethod
    def from_run(
        cls,
        run: Run,
        name: str | None = None,
        queries: Mapping[str, str] | None = None,
        dtype: np.dtype = np.dtype(np.float32),
    ) -> "Ranking":
        """Create a Ranking object from a TREC run.

        :param run: TREC run.
        :param name: Method name.
        :param queries: Query IDs mapped to queries.
        :param dtype: How the score should be represented in the data frame.
        :return: The resulting ranking.
        """
        df = pd.DataFrame.from_dict(dict(run)).stack().reset_index()
        df.columns = ("id", "q_id", "score")
        return cls(df, name=name, queries=queries, dtype=dtype, copy=False)

    @classmethod
    def from_file(
        cls,
        f: "Path",
        queries: Mapping[str, str] | None = None,
        dtype: np.dtype = np.dtype(np.float32),
    ) -> "Ranking":
        """Create a Ranking object from a runfile in TREC format.

        :param f: TREC runfile to read.
        :param queries: Query IDs mapped to queries.
        :param dtype: How the score should be represented in the data frame.
        :return: The resulting ranking.
        """
        df = pd.read_csv(
            f,
            sep=r"\s+",
            skipinitialspace=True,
            header=None,
            names=["q_id", "q0", "id", "rank", "score", "name"],
        )
        return cls(df, name=df["name"][0], queries=queries, dtype=dtype, copy=False)



import abc
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Sequence

    import numpy as np


class Encoder(abc.ABC):
    """Base class for encoders."""

    @abc.abstractmethod
    def _encode(self, texts: "Sequence[str]") -> "np.ndarray":
        pass

    def __call__(self, texts: "Sequence[str]") -> "np.ndarray":
        """Encode a list of texts.

        :param texts: The texts to encode.
        :return: The resulting vector representations.
        """
        return self._encode(texts)



""".. include:: ../docs/pyterrier.md"""  # noqa: D400, D415

from typing import TYPE_CHECKING

import pyterrier as pt

from fast_forward.ranking import Ranking

if TYPE_CHECKING:
    import pandas as pd

    from fast_forward.index.base import Index


class FFScore(pt.Transformer):
    """PyTerrier transformer that computes scores using a Fast-Forward index."""

    def __init__(self, index: "Index") -> None:
        """Create an FFScore transformer.

        :param index: The Fast-Forward index.
        """
        self._index = index
        super().__init__()

    def transform(self, inp: "pd.DataFrame") -> "pd.DataFrame":
        """Compute the scores for all query-document pairs in the data frame.

        The previous scores are moved to the "score_0" column.

        :param inp: The PyTerrier data frame.
        :return: A data frame with the computed scores.
        """
        ff_scores = self._index(
            Ranking(
                inp.rename(columns={"qid": "q_id", "docno": "id"}),
                copy=False,
                # since this ranking is only used for scoring, sorting is not necessary
                is_sorted=True,
            )
        )._df.rename(columns={"q_id": "qid", "id": "docno"})

        result = ff_scores[["qid", "docno", "score", "query"]].merge(
            inp[["qid", "docno", "score"]],
            on=["qid", "docno"],
            suffixes=(None, "_0"),
        )
        return pt.model.add_ranks(result, single_query=False)

    def __repr__(self) -> str:
        """Return a string representation.

        The representation is unique w.r.t. the index and its query encoder.

        :return: The representation.
        """
        cls_name = self.__class__.__name__
        index_id = id(self._index)
        encoder_id = id(self._index._query_encoder)
        return f"{cls_name}({index_id}, {encoder_id})"


class FFInterpolate(pt.Transformer):
    """PyTerrier transformer that interpolates scores computed by `FFScore`."""

    def __init__(self, alpha: float) -> None:
        """Create an FFInterpolate transformer.

        :param alpha: The interpolation parameter.
        """
        # attribute name needs to be exactly this for pyterrier.GridScan to work
        self.alpha = alpha
        super().__init__()

    def transform(self, inp: "pd.DataFrame") -> "pd.DataFrame":
        """Interpolate the scores as `alpha * score_0 + (1 - alpha) * score`.

        :param inp: The PyTerrier data frame.
        :return: A data frame with the interpolated scores.
        """
        new_df = inp[["qid", "docno", "query"]].copy()
        new_df["score"] = self.alpha * inp["score_0"] + (1 - self.alpha) * inp["score"]
        return pt.model.add_ranks(new_df, single_query=False)



""".. include:: ../docs/index.md"""  # noqa: D400, D415

from fast_forward.index.base import Index, Mode
from fast_forward.index.disk import OnDiskIndex
from fast_forward.index.memory import InMemoryIndex

__all__ = ["Index", "Mode", "OnDiskIndex", "InMemoryIndex"]



from collections import defaultdict
from typing import TYPE_CHECKING

import numpy as np

from fast_forward.index.base import Mode

if TYPE_CHECKING:
    from collections.abc import Iterable


def get_indices(
    ids: "Iterable[str]",
    mode: Mode,
    doc_id_to_idx: dict[str, list[int]],
    psg_id_to_idx: dict[str, int],
) -> tuple[list[int], list[str]]:
    """Find the array indices for a list of document/passage IDs.

    :param ids: IDs to return vectors for.
    :param mode: The ranking mode.
    :param doc_id_to_idx: Document IDs mapped to non-chunked indices.
    :param psg_id_to_idx: Passage IDs mapped to non-chunked indices.
    :raises IndexError: When the ID cannot be found in the index.
    :return: The indices and corresponding IDs.
    """
    indices = []
    ids_ = []
    for id in ids:
        if mode in (Mode.MAXP, Mode.AVEP):
            cur_indices = doc_id_to_idx.get(id, [])
        elif mode == Mode.FIRSTP:
            cur_indices = doc_id_to_idx.get(id, [])[:1]
        else:
            psg_idx = psg_id_to_idx.get(id)
            cur_indices = [] if psg_idx is None else [psg_idx]

        if len(cur_indices) == 0:
            raise IndexError(f"ID {id} not found in the index.")
        indices.extend(cur_indices)
        ids_.extend([id] * len(cur_indices))
    return indices, ids_


class ChunkIndexer:
    """Utility class for retrieving vectors from chunked indexes."""

    def __init__(
        self,
        chunks: list[np.ndarray] | list[np.memmap],
        doc_id_to_idx: dict[str, list[int]],
        psg_id_to_idx: dict[str, int],
    ) -> None:
        """Create a chunk indexer.

        The first chunk may have a different size than the others.

        :param chunks: A list of chunks.
        :param doc_id_to_idx: Document IDs mapped to non-chunked indices.
        :param psg_id_to_idx: Passage IDs mapped to non-chunked indices.
        """
        self._chunks = chunks
        self._doc_id_to_idx = doc_id_to_idx
        self._psg_id_to_idx = psg_id_to_idx

    def _get_chunk_indices(self, idx: int) -> tuple[int, int]:
        """Given an index, compute the chunk index and index within that chunk.

        The input index may be in `[0, N]`, where `N` is the total number of vectors.

        :param idx: The input index.
        :return: Chunk index and index within that chunk.
        """
        # the first chunk might be larger
        if idx < self._chunks[0].shape[0]:
            return 0, idx

        idx_ = idx - self._chunks[0].shape[0]
        return (
            int(idx_ / self._chunks[1].shape[0]) + 1,
            idx_ % self._chunks[1].shape[0],
        )

    def __call__(
        self, ids: "Iterable[str]", mode: Mode
    ) -> tuple[np.ndarray, list[str]]:
        """Retrieve vectors for the given IDs from the chunks.

        :param ids: IDs to return vectors for.
        :raises IndexError: When the ID cannot be found in the index.
        :return: The vectors and corresponding IDs.
        """
        indices, ids_ = get_indices(ids, mode, self._doc_id_to_idx, self._psg_id_to_idx)

        if len(indices) == 0:
            return np.array([]), []
        if len(self._chunks) == 1:
            return self._chunks[0][indices], ids_

        # group indexes by chunk and index each chunk individually
        items_by_chunk = defaultdict(lambda: ([], []))
        for idx, id in zip(indices, ids_):
            chunk_idx, idx_in_chunk = self._get_chunk_indices(idx)
            items_by_chunk[chunk_idx][0].append(idx_in_chunk)
            items_by_chunk[chunk_idx][1].append(id)

        result = []
        out_ids = []
        for chunk_idx, (idx_in_chunk, ids_in_chunk) in items_by_chunk.items():
            result.append(self._chunks[chunk_idx][idx_in_chunk])
            out_ids.extend(ids_in_chunk)

        return np.concatenate(result), out_ids



import unittest

import numpy as np

from fast_forward.encoder import LambdaEncoder
from fast_forward.encoder.transformer import (
    TCTColBERTDocumentEncoder,
    TCTColBERTQueryEncoder,
    TASBEncoder,
    BGEEncoder,
    ContrieverEncoder,
)

from ._constants import (
    TCT_COLBERT_QUERY_EXPECTED,
    TCT_COLBERT_DOCUMENT_EXPECTED,
    TAS_B_EXPECTED,
    CONTRIEVER_EXPECTED,
    BGE_ENCODER_EXPECTED,
)

TEST_INPUTS = ["input 1", "second input", "3rd input " * 100]


class TestLambdaEncoder(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.encoder = LambdaEncoder(lambda q: np.zeros(shape=(768,)))

    def test_encoder(self):
        np.testing.assert_equal(self.encoder(TEST_INPUTS), np.zeros(shape=(3, 768)))


class TestTCTColBERTEncoder(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.query_encoder = TCTColBERTQueryEncoder()
        cls.doc_encoder = TCTColBERTDocumentEncoder()

    def test_query_encoder(self):
        np.testing.assert_almost_equal(
            self.query_encoder(TEST_INPUTS),
            TCT_COLBERT_QUERY_EXPECTED,
            decimal=5,
        )

    def test_doc_encoder(self):
        np.testing.assert_almost_equal(
            self.doc_encoder(TEST_INPUTS),
            TCT_COLBERT_DOCUMENT_EXPECTED,
            decimal=5,
        )


class TestTASBEncoder(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.encoder = TASBEncoder()

    def test_encoder(self):
        np.testing.assert_almost_equal(
            self.encoder(TEST_INPUTS),
            TAS_B_EXPECTED,
            decimal=5,
        )


class TestContrieverEncoder(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.encoder = ContrieverEncoder()

    def test_encoder(self):
        np.testing.assert_almost_equal(
            self.encoder(TEST_INPUTS),
            CONTRIEVER_EXPECTED,
            decimal=5,
        )


class TestBGEEncoder(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.encoder = BGEEncoder()

    def test_encoder(self):
        np.testing.assert_almost_equal(
            self.encoder(TEST_INPUTS),
            BGE_ENCODER_EXPECTED,
            decimal=5,
        )


if __name__ == "__main__":
    unittest.main()



import logging
from collections import defaultdict
from typing import TYPE_CHECKING

import numpy as np
from tqdm import tqdm

from fast_forward.index.base import IDSequence, Index, Mode
from fast_forward.index.util import ChunkIndexer

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

    from fast_forward.encoder.base import Encoder
    from fast_forward.quantizer import Quantizer

LOGGER = logging.getLogger(__name__)


class InMemoryIndex(Index):
    """Fast-Forward index that is held entirely in memory."""

    def __init__(
        self,
        query_encoder: "Encoder | None" = None,
        quantizer: "Quantizer | None" = None,
        mode: Mode = Mode.MAXP,
        encoder_batch_size: int = 32,
        init_size: int = 2**16,
        alloc_size: int = 2**16,
    ) -> None:
        """Create an index in memory.

        :param query_encoder: The query encoder to use.
        :param quantizer: The quantizer to use.
        :param mode: The ranking mode.
        :param encoder_batch_size: Batch size for the query encoder.
        :param init_size: Size of initial chunk (number of vectors).
        :param alloc_size: Size of additionally allocated chunks (number of vectors).
        """
        self._chunks = []
        self._init_size = init_size
        self._alloc_size = alloc_size
        self._idx_in_cur_chunk = 0
        self._doc_id_to_idx = defaultdict(list)
        self._psg_id_to_idx = {}
        self._chunk_indexer = ChunkIndexer(
            self._chunks, self._doc_id_to_idx, self._psg_id_to_idx
        )

        super().__init__(
            query_encoder=query_encoder,
            quantizer=quantizer,
            mode=mode,
            encoder_batch_size=encoder_batch_size,
        )

    def _get_num_vectors(self) -> int:
        # account for the fact that the first chunk might be larger
        if len(self._chunks) < 2:
            return self._idx_in_cur_chunk
        return (
            self._chunks[0].shape[0]
            + (len(self._chunks) - 2) * self._alloc_size
            + self._idx_in_cur_chunk
        )

    def _get_internal_dim(self) -> int | None:
        if len(self._chunks) > 0:
            return self._chunks[0].shape[-1]
        return None

    def _add(
        self,
        vectors: np.ndarray,
        doc_ids: IDSequence,
        psg_ids: IDSequence,
    ) -> None:
        # if this is the first call to _add, no chunks exist
        if len(self._chunks) == 0:
            self._chunks.append(
                np.zeros((self._init_size, vectors.shape[-1]), dtype=vectors.dtype)
            )

        # assign passage and document IDs
        for i, doc_id in enumerate(doc_ids, len(self)):
            if doc_id is not None:
                self._doc_id_to_idx[doc_id].append(i)

        for i, psg_id in enumerate(psg_ids, len(self)):
            if psg_id is None:
                continue
            if psg_id in self._psg_id_to_idx:
                raise RuntimeError(f"Passage ID {psg_id} already exists.")
            self._psg_id_to_idx[psg_id] = i

        # add vectors to chunks
        added = 0
        num_vectors = vectors.shape[0]
        while added < num_vectors:
            cur_chunk_size, dim = self._chunks[-1].shape

            # if current chunk is full, add a new one
            if self._idx_in_cur_chunk == cur_chunk_size:
                LOGGER.debug("adding new chunk")
                self._chunks.append(np.zeros((self._alloc_size, dim)))
                self._idx_in_cur_chunk = 0
                cur_chunk_size = self._alloc_size

            to_add = min(
                num_vectors - added,
                cur_chunk_size,
                cur_chunk_size - self._idx_in_cur_chunk,
            )
            self._chunks[-1][
                self._idx_in_cur_chunk : self._idx_in_cur_chunk + to_add
            ] = vectors[added : added + to_add]
            added += to_add
            self._idx_in_cur_chunk += to_add

    def consolidate(self) -> None:
        """Combine all chunks of the index in one contiguous section in the memory."""
        # combine all chunks up to the last one entirely, and take only whats in use of
        # the last one
        self._chunks = [
            np.concatenate(
                self._chunks[:-1] + [self._chunks[-1][: self._idx_in_cur_chunk]]
            )
        ]
        self._idx_in_cur_chunk = self._chunks[0].shape[0]
        self._chunk_indexer._chunks = self._chunks

    def _get_doc_ids(self) -> set[str]:
        return set(self._doc_id_to_idx.keys())

    def _get_psg_ids(self) -> set[str]:
        return set(self._psg_id_to_idx.keys())

    def _get_vectors(self, ids: "Iterable[str]") -> tuple[np.ndarray, list[str]]:
        return self._chunk_indexer(ids, self.mode)

    def _batch_iter(
        self, batch_size: int
    ) -> "Iterator[tuple[np.ndarray, IDSequence, IDSequence]]":
        LOGGER.info("creating ID mappings for this index")
        idx_to_doc_id = {
            idx: doc_id
            for doc_id, idxs in tqdm(self._doc_id_to_idx.items())
            for idx in idxs
        }
        idx_to_psg_id = {
            idx: psg_id for psg_id, idx in tqdm(self._psg_id_to_idx.items())
        }

        num_vectors = len(self)
        for i in range(0, num_vectors, batch_size):
            j = min(i + batch_size, num_vectors)

            # the current batch is between i (incl.) and j (excl.)
            i_chunk_idx, i_idx_in_chunk = self._chunk_indexer._get_chunk_indices(i)
            j_chunk_idx, j_idx_in_chunk = self._chunk_indexer._get_chunk_indices(j - 1)

            arrays = []

            # if the batch spans multiple chunks, collect them in a list
            while i_chunk_idx < j_chunk_idx:
                arrays.append(self._chunks[i_chunk_idx][i_idx_in_chunk:])
                i_chunk_idx += 1
                i_idx_in_chunk = 0

            # now i_chunk_idx == j_chunk_idx
            arrays.append(
                self._chunks[i_chunk_idx][i_idx_in_chunk : j_idx_in_chunk + 1]
            )

            yield (
                np.concatenate(arrays),
                list(map(idx_to_doc_id.get, range(i, j))),
                list(map(idx_to_psg_id.get, range(i, j))),
            )



""".. include:: ../docs/encoder.md"""  # noqa: D400, D415

from typing import TYPE_CHECKING

import numpy as np

from fast_forward.encoder.base import Encoder
from fast_forward.encoder.transformer import (
    BGEEncoder,
    ContrieverEncoder,
    TASBEncoder,
    TCTColBERTDocumentEncoder,
    TCTColBERTQueryEncoder,
    TransformerEncoder,
)

if TYPE_CHECKING:
    from collections.abc import Callable, Sequence

__all__ = [
    "Encoder",
    "LambdaEncoder",
    "TransformerEncoder",
    "TCTColBERTQueryEncoder",
    "TCTColBERTDocumentEncoder",
    "TASBEncoder",
    "ContrieverEncoder",
    "BGEEncoder",
]


class LambdaEncoder(Encoder):
    """Encoder adapter class for arbitrary encoding functions."""

    def __init__(self, f: "Callable[[str], np.ndarray]") -> None:
        """Create a lambda encoder.

        :param f: Function to encode a single piece of text.
        """
        super().__init__()
        self._f = f

    def _encode(self, texts: "Sequence[str]") -> np.ndarray:
        return np.array(list(map(self._f, texts)))



import abc
import logging
from collections.abc import Iterable, Iterator, Sequence
from enum import Enum
from time import perf_counter

import numpy as np
import pandas as pd
from tqdm import tqdm

from fast_forward.encoder.base import Encoder
from fast_forward.quantizer import Quantizer
from fast_forward.ranking import Ranking

LOGGER = logging.getLogger(__name__)


class Mode(Enum):
    """Enum used to set the ranking mode of an index."""

    PASSAGE = 1
    MAXP = 2
    FIRSTP = 3
    AVEP = 4


IDSequence = Sequence[str | None]


class Index(abc.ABC):
    """Abstract base class for Fast-Forward indexes."""

    _query_encoder: Encoder | None = None
    _quantizer: Quantizer | None = None

    def __init__(
        self,
        query_encoder: Encoder | None = None,
        quantizer: Quantizer | None = None,
        mode: Mode = Mode.MAXP,
        encoder_batch_size: int = 32,
    ) -> None:
        """Create an index.

        :param query_encoder: The query encoder to use.
        :param quantizer: The quantizer to use.
        :param mode: The ranking mode.
        :param encoder_batch_size: The encoder batch size.
        """
        super().__init__()
        if query_encoder is not None:
            self.query_encoder = query_encoder
        self.mode = mode
        if quantizer is not None:
            self.quantizer = quantizer
        self._encoder_batch_size = encoder_batch_size

    def encode_queries(self, queries: Sequence[str]) -> np.ndarray:
        """Encode queries.

        :param queries: The queries to encode.
        :raises RuntimeError: When no query encoder exists.
        :return: The query representations.
        """
        if self.query_encoder is None:
            raise RuntimeError("Index does not have a query encoder.")

        result = []
        for i in range(0, len(queries), self._encoder_batch_size):
            batch = queries[i : i + self._encoder_batch_size]
            result.append(self.query_encoder(batch))
        return np.concatenate(result)

    @property
    def query_encoder(self) -> Encoder | None:
        """Return the query encoder if it exists.

        :return: The query encoder (if any).
        """
        return self._query_encoder

    @query_encoder.setter
    def query_encoder(self, encoder: Encoder) -> None:
        """Set the query encoder.

        :param encoder: The query encoder to set.
        """
        assert isinstance(encoder, Encoder)
        self._query_encoder = encoder

    @property
    def quantizer(self) -> Quantizer | None:
        """Return the quantizer if it exists.

        :return: The quantizer (if any).
        """
        return self._quantizer

    def _on_quantizer_set(self) -> None:
        """Handle a quantizer being attached to this index."""
        pass

    @quantizer.setter
    def quantizer(self, quantizer: Quantizer) -> None:
        """Set the quantizer.

        This is only possible before any vectors are added to the index.

        :param quantizer: The new quantizer.
        :raises RuntimeError: When the index is not empty.
        """
        assert isinstance(quantizer, Quantizer)

        if len(self) > 0:
            raise RuntimeError("Quantizers can only be attached to empty indexes.")
        self._quantizer = quantizer
        self._on_quantizer_set()
        quantizer.set_attached()

    @property
    def mode(self) -> Mode:
        """Return the ranking mode.

        :return: The ranking mode.
        """
        return self._mode

    @mode.setter
    def mode(self, mode: Mode) -> None:
        """Set the ranking mode.

        :param mode: The ranking mode to set.
        """
        assert isinstance(mode, Mode)
        self._mode = mode

    @abc.abstractmethod
    def _get_internal_dim(self) -> int | None:
        pass

    @property
    def dim(self) -> int | None:
        """Return the dimensionality of the vector index.

        If no vectors exist, return `None`.
        If a quantizer is used, return the dimension of the codes.

        :return: The dimensionality (if any).
        """
        if self._quantizer is not None:
            return self._quantizer.dims[0]
        return self._get_internal_dim()

    @abc.abstractmethod
    def _get_doc_ids(self) -> set[str]:
        pass

    @property
    def doc_ids(self) -> set[str]:
        """Return all unique document IDs.

        :return: The document IDs.
        """
        return self._get_doc_ids()

    @abc.abstractmethod
    def _get_psg_ids(self) -> set[str]:
        pass

    @property
    def psg_ids(self) -> set[str]:
        """Return all unique passage IDs.

        :return: The passage IDs.
        """
        return self._get_psg_ids()

    @abc.abstractmethod
    def _get_num_vectors(self) -> int:
        pass

    def __len__(self) -> int:
        """Return the index size.

        :return: The number of vectors in the index.
        """
        return self._get_num_vectors()

    @abc.abstractmethod
    def _add(
        self,
        vectors: np.ndarray,
        doc_ids: IDSequence,
        psg_ids: IDSequence,
    ) -> None:
        """Add vector representations and corresponding IDs to the index.

        Document IDs may have duplicates, passage IDs are assumed to be unique.
        Vectors may be quantized.

        Specific to index implementation.

        :param vectors:
            The representations, shape `(num_vectors, dim)` or
            `(num_vectors, quantized_dim)`.
        :param doc_ids: The corresponding document IDs.
        :param psg_ids: The corresponding passage IDs.
        """
        pass

    def add(
        self,
        vectors: np.ndarray,
        doc_ids: IDSequence | None = None,
        psg_ids: IDSequence | None = None,
    ) -> None:
        """Add vector representations and corresponding IDs to the index.

        Only one of `doc_ids` and `psg_ids` may be `None`.
        Individual IDs in the sequence may also be `None`, but each vector must have at
        least one associated ID.
        Document IDs may have duplicates, passage IDs must be unique.

        :param vectors: The representations, shape `(num_vectors, dim)`.
        :param doc_ids: The corresponding document IDs (may be duplicate).
        :param psg_ids: The corresponding passage IDs (must be unique).
        :raises ValueError: When the number of IDs does not match the number of vectors.
        :raises ValueError:
            When the input vector and index dimensionalities do not match.
        :raises ValueError: When a vector has neither a document nor a passage ID.
        :raises RuntimeError: When items can't be added to the index for any reason.
        """
        num_vectors, dim = vectors.shape

        if doc_ids is None:
            doc_ids = [None] * num_vectors
        if psg_ids is None:
            psg_ids = [None] * num_vectors
        if not len(doc_ids) == len(psg_ids) == num_vectors:
            raise ValueError("Number of IDs does not match number of vectors.")

        if self.dim is not None and dim != self.dim:
            raise ValueError(
                f"Input vector dimensionality ({dim}) does not match "
                f"index dimensionality ({self.dim})."
            )

        for doc_id, psg_id in zip(doc_ids, psg_ids):
            if doc_id is None and psg_id is None:
                raise ValueError("Vector has neither document nor passage ID.")

        self._add(
            vectors if self.quantizer is None else self.quantizer.encode(vectors),
            doc_ids,
            psg_ids,
        )

    @abc.abstractmethod
    def _get_vectors(self, ids: Iterable[str]) -> tuple[np.ndarray, list[str]]:
        """Get vectors and corresponding IDs from the index.

        Return a tuple containing:
            * A single array of all vectors necessary to compute the scores for each
                document/passage.
            * A list of document/passage IDs to identify each vector in the array.

        The output of this function depends on the current mode.

        If a quantizer is used, this function returns quantized vectors.

        Specific to index implementation.

        :param ids: The document/passage IDs to get the representations for.
        :raises IndexError: When a requested ID is not found in the index.
        :return: The vectors and corresponding indices.
        """
        pass

    def _compute_scores(
        self, data: pd.DataFrame | pd.Series, query_vectors: np.ndarray
    ) -> pd.DataFrame:
        """Compute scores for a data frame.

        The input data frame needs a "q_no" column with unique query numbers.

        :param data: Input data frame (or series).
        :param query_vectors: All query vectors indexed by "q_no".
        :return: Data frame with computed scores.
        """
        # get all required vectors and corresponding IDs from the FF index
        vectors, vec_ids = self._get_vectors(data["id"].unique())
        if self.quantizer is not None:
            vectors = self.quantizer.decode(vectors)

        # merge data frames so "id_idx" can be used to index arrays
        df_vec_ids = pd.DataFrame(vec_ids, columns=["id"])
        df_vec_ids["id_idx"] = df_vec_ids.index
        df_merged = df_vec_ids.merge(data[["id", "q_no"]], how="left", on="id")

        # compute all dot products (scores)
        q_reps = query_vectors[df_merged["q_no"].tolist()]
        d_reps = vectors[df_merged["id_idx"].tolist()]
        df_merged["ff_score"] = np.sum(q_reps * d_reps, axis=1)

        # select aggregation operation based on current mode
        if self.mode == Mode.MAXP:
            op = "max"
        elif self.mode == Mode.AVEP:
            op = "mean"
        else:
            op = "first"
        df_agg = df_merged.groupby(["id", "q_no"], as_index=False).aggregate(op)

        return data.merge(df_agg, on=["id", "q_no"])

    def _early_stopping(
        self,
        df: pd.DataFrame,
        query_vectors: np.ndarray,
        cutoff: int,
        alpha: float,
        depths: Iterable[int],
    ) -> pd.DataFrame:
        """Compute scores with early stopping for a data frame.

        The input data frame needs a "q_no" column with unique query numbers.

        :param df: Input data frame.
        :param query_vectors: All query vectors indexed by "q_no".
        :param cutoff: Cut-off depth for early stopping.
        :param alpha: Interpolation parameter.
        :param depths: Depths to compute scores at.
        :return: Data frame with computed scores.
        """
        # data frame for computed scores
        scores_so_far = pd.DataFrame()

        # [a, b] is the interval for which the scores are computed in each step
        a = 0
        for b in sorted(depths):
            if b < cutoff:
                continue

            # identify queries which do not meet the early stopping criterion
            if a == 0:
                # first iteration: take all queries
                q_ids_left = pd.unique(df["q_id"])
            else:
                # subsequent iterations: compute ES criterion
                q_ids_left = (
                    scores_so_far.groupby("q_id")
                    .filter(
                        lambda g: g["int_score"].nlargest(cutoff).iat[-1]
                        < alpha * g["score"].iat[-1] + (1 - alpha) * g["ff_score"].max()
                    )["q_id"]
                    .drop_duplicates()
                    .to_list()
                )
            LOGGER.info("depth %s: %s queries left", b, len(q_ids_left))

            # take the next chunk with b-a docs/passages for each query
            chunk = (
                df.loc[df["q_id"].isin(q_ids_left)]
                .groupby("q_id")
                .nth(list(range(a, b)))
            )

            # stop if no pairs are left
            if len(chunk) == 0:
                break

            # compute scores for the chunk and merge
            out = self._compute_scores(chunk, query_vectors)[["orig_index", "ff_score"]]
            chunk_scores = chunk.merge(out, on="orig_index", suffixes=(None, "_"))

            # compute interpolated scores
            chunk_scores["int_score"] = (
                alpha * chunk_scores["score"] + (1 - alpha) * chunk_scores["ff_score"]
            )

            scores_so_far = pd.concat(
                [scores_so_far, chunk_scores],
                axis=0,
            )

            a = b
        return scores_so_far.join(df, on="orig_index", lsuffix="", rsuffix="_")

    def __call__(
        self,
        ranking: Ranking,
        early_stopping: int | None = None,
        early_stopping_alpha: float | None = None,
        early_stopping_depths: Iterable[int] | None = None,
        batch_size: int | None = None,
    ) -> Ranking:
        """Compute scores for a ranking.

        :param ranking: The ranking to compute scores for. Must have queries attached.
        :param early_stopping: Perform early stopping at this cut-off depth.
        :param early_stopping_alpha: Interpolation parameter for early stopping.
        :param early_stopping_depths: Depths for early stopping.
        :param batch_size: How many queries to process at once.
        :raises ValueError: When the ranking has no queries attached.
        :raises ValueError: When early stopping is enabled but arguments are missing.
        :raises IndexError: When an ID in the ranking is not found in the index.
        :return: Ranking with the computed scores.
        """
        if not ranking.has_queries:
            raise ValueError("Input ranking has no queries attached.")
        if early_stopping is not None and (
            early_stopping_alpha is None or early_stopping_depths is None
        ):
            raise ValueError("Early stopping requires alpha and depths.")
        t0 = perf_counter()

        # get all unique queries and query IDs and map to unique numbers (0 to m)
        query_df = (
            ranking._df[["q_id", "query"]].drop_duplicates().reset_index(drop=True)
        )
        query_df["q_no"] = query_df.index
        df_with_q_no = ranking._df.merge(query_df, on="q_id", suffixes=(None, "_"))

        # early stopping splits the data frame, hence we need to keep track of the
        # original index
        df_with_q_no["orig_index"] = df_with_q_no.index

        # batch encode queries
        query_vectors = self.encode_queries(list(query_df["query"]))

        def _get_result(df: pd.DataFrame) -> pd.DataFrame:
            if early_stopping is None:
                return self._compute_scores(df, query_vectors)

            assert early_stopping_alpha is not None
            assert early_stopping_depths is not None
            return self._early_stopping(
                df,
                query_vectors,
                early_stopping,
                early_stopping_alpha,
                early_stopping_depths,
            )

        num_queries = len(query_df)
        if batch_size is None or batch_size >= num_queries:
            result = _get_result(df_with_q_no)
        else:
            # assign batch indices to query IDs
            df_with_q_no["batch_idx"] = (
                df_with_q_no.groupby("q_id").ngroup() / batch_size
            ).astype(int)

            chunks = []
            num_batches = int(num_queries / batch_size) + 1
            for batch_idx in tqdm(range(num_batches)):
                batch = df_with_q_no[df_with_q_no["batch_idx"] == batch_idx]
                chunks.append(_get_result(batch))
            result = pd.concat(chunks)

        result["score"] = result["ff_score"]
        LOGGER.info("computed scores in %s seconds", perf_counter() - t0)
        return Ranking(
            result,
            name="fast-forward",
            dtype=ranking._df.dtypes["score"],
            copy=False,
            is_sorted=False,
        )

    @abc.abstractmethod
    def _batch_iter(
        self, batch_size: int
    ) -> Iterator[tuple[np.ndarray, IDSequence, IDSequence]]:
        """Iterate over the index in batches.

        If a quantizer is used, the vectors are the quantized codes.
        When an ID does not exist, it must be set to `None`.

        Specific to index implementation.

        :param batch_size: The batch size.
        :return: Iterator yielding vectors, document IDs, passage IDs (in batches).
        """
        pass

    def batch_iter(
        self, batch_size: int
    ) -> Iterator[tuple[np.ndarray, IDSequence, IDSequence]]:
        """Iterate over all vectors, document IDs, and passage IDs in batches.

        IDs may be either strings or `None`.

        :param batch_size: Batch size.
        :return: Iterator yielding batches of vectors, document IDs (if any),
            passage IDs (if any).
        """
        if self._quantizer is None:
            yield from self._batch_iter(batch_size)

        else:
            for batch in self._batch_iter(batch_size):
                yield self._quantizer.decode(batch[0]), batch[1], batch[2]

    def __iter__(
        self,
    ) -> Iterator[tuple[np.ndarray, str | None, str | None]]:
        """Iterate over all vectors, document IDs, and passage IDs.

        :return: Iterator yielding vectors, document IDs (if any), passage IDs (if any).
        """
        for vectors, doc_ids, psg_ids in self.batch_iter(2**9):
            yield from zip(vectors, doc_ids, psg_ids)



import unittest

import numpy as np

from fast_forward.encoder import LambdaEncoder
from fast_forward.index.memory import InMemoryIndex
from fast_forward.quantizer.nanopq import NanoPQ
from fast_forward.util.indexer import Indexer


class TestIndexer(unittest.TestCase):
    def setUp(self):
        self.target_index = InMemoryIndex()
        self.indexer = Indexer(
            self.target_index,
            LambdaEncoder(lambda q: np.zeros(shape=(16,))),
            encoder_batch_size=2,
            batch_size=4,
        )

    def test_from_dicts(self):
        dicts = [
            {"text": "123", "doc_id": "d1", "psg_id": "d1_p1"},
            {"text": "234", "doc_id": "d1", "psg_id": "d1_p2"},
            {"text": "456", "doc_id": "d1", "psg_id": "d1_p3"},
            {"text": "567", "doc_id": "d2", "psg_id": "d2_p1"},
            {"text": "678", "doc_id": "d3", "psg_id": "d3_p1"},
            {"text": "890", "doc_id": "d4"},
            {"text": "901", "psg_id": "d5_p1"},
        ]
        self.indexer.from_dicts(dicts)
        self.assertEqual(7, len(self.target_index))
        self.assertEqual(set(("d1", "d2", "d3", "d4")), self.target_index.doc_ids)
        self.assertEqual(
            set(("d1_p1", "d1_p2", "d1_p3", "d2_p1", "d3_p1", "d5_p1")),
            self.target_index.psg_ids,
        )

        with self.assertRaises(RuntimeError):
            Indexer(self.target_index, encoder=None).from_dicts(dicts)

    def test_from_index(self):
        source_index = InMemoryIndex()
        source_index.add(
            np.zeros((16, 16), dtype=np.float32), doc_ids=[f"d{i}" for i in range(16)]
        )
        self.indexer.from_index(source_index)
        self.assertEqual(source_index.doc_ids, self.target_index.doc_ids)
        self.assertEqual(16, len(self.target_index))

    def test_quantizer(self):
        for quantizer_fit_batches in (1, 2):
            target_index = InMemoryIndex()
            indexer = Indexer(
                target_index,
                encoder=LambdaEncoder(
                    lambda q: np.random.normal(size=(32,)).astype(np.float32)
                ),
                quantizer=NanoPQ(4, 8),
                batch_size=16,
                quantizer_fit_batches=quantizer_fit_batches,
            )
            indexer.from_dicts(
                [{"text": f"text_{i}", "doc_id": f"d{i}"} for i in range(64)]
            )
            self.assertTrue(target_index.quantizer._trained)

        with self.assertRaises(ValueError):
            quantizer = NanoPQ(4, 8)
            quantizer.fit(np.random.normal(size=(64, 64)).astype(np.float32))
            Indexer(self.target_index, quantizer=quantizer)

        with self.assertRaises(ValueError):
            self.target_index.add(np.zeros(shape=(8, 16), dtype=np.float32))
            Indexer(self.target_index, quantizer=NanoPQ(4, 8))


if __name__ == "__main__":
    unittest.main()



import logging
from collections import defaultdict
from typing import TYPE_CHECKING, cast

import h5py
import numpy as np
from tqdm import tqdm

import fast_forward
from fast_forward.index.base import IDSequence, Index, Mode
from fast_forward.index.memory import InMemoryIndex
from fast_forward.index.util import ChunkIndexer, get_indices
from fast_forward.quantizer import Quantizer

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator
    from pathlib import Path

    from fast_forward.encoder.base import Encoder

LOGGER = logging.getLogger(__name__)


# h5py does not play nice with pyright, so we add lots of ignores in this class
class OnDiskIndex(Index):
    """Fast-Forward index that is read on-demand from disk (HDF5 format).

    The `max_indexing_size` argument works around an
    [h5py limitation](https://docs.h5py.org/en/latest/high/dataset.html#fancy-indexing).

    If `memory_mapped` is set to `True`, the vectors are accessed using memory maps,
    which is usually faster. In this case, `max_indexing_size` has no effect.
    """

    def __init__(
        self,
        index_file: "Path",
        query_encoder: "Encoder | None" = None,
        quantizer: Quantizer | None = None,
        mode: Mode = Mode.MAXP,
        encoder_batch_size: int = 32,
        init_size: int = 2**16,
        chunk_size: int = 2**16,
        max_id_length: int = 8,
        overwrite: bool = False,
        memory_mapped: bool = False,
        max_indexing_size: int = 2**10,
    ) -> None:
        """Create an index on disk.

        :param index_file: The index file to create (or overwrite).
        :param query_encoder: The query encoder.
        :param quantizer: The quantizer to use.
        :param mode: The ranking mode.
        :param encoder_batch_size: Batch size for the query encoder.
        :param init_size: Initial size to allocate (number of vectors).
        :param chunk_size: Size of chunks (HDF5).
        :param max_id_length:
            Maximum length of document and passage IDs (number of characters).
        :param overwrite: Overwrite index file if it exists.
        :param memory_mapped: Use memory maps for retrieval of vectors.
        :param max_indexing_size:
            Maximum number of vectors to retrieve from the HDF5 dataset at once.
        :raises ValueError: When the file exists and `overwrite=False`.
        """
        if index_file.exists() and not overwrite:
            raise ValueError(f"File {index_file} exists.")

        self._index_file = index_file.absolute()
        self._doc_id_to_idx = defaultdict(list)
        self._psg_id_to_idx = {}

        self._init_size = init_size
        self._chunk_size = chunk_size
        self._max_id_length = max_id_length
        self._memory_mapped = memory_mapped
        self._max_indexing_size = max_indexing_size

        # the memory maps are created on-demand in _get_vectors
        self._mmap_indexer = None

        LOGGER.debug("creating file %s", self._index_file)
        with h5py.File(self._index_file, "w") as fp:
            fp.attrs["num_vectors"] = 0
            fp.attrs["ff_version"] = fast_forward.__version__

        super().__init__(
            query_encoder=query_encoder,
            quantizer=quantizer,
            mode=mode,
            encoder_batch_size=encoder_batch_size,
        )

    def _get_mmap_indexer(self) -> ChunkIndexer:
        """Create and return a chunk indexer using memory-mapped arrays.

        :raises RuntimeError: When the dataset is not chunked in the required way.
        :return: The chunk indexer.
        """
        if self._mmap_indexer is None:
            with h5py.File(self._index_file, "r") as fp:
                if (
                    fp["vectors"].chunks is None  # pyright: ignore[reportAttributeAccessIssue]
                    or fp["vectors"].chunks[1] != fp["vectors"].shape[1]  # pyright: ignore[reportAttributeAccessIssue]
                ):
                    raise RuntimeError("This index does not support memory maps.")
                arrays = [
                    np.memmap(
                        self._index_file,
                        mode="r",
                        shape=fp["vectors"].chunks,  # pyright: ignore[reportAttributeAccessIssue]
                        offset=fp["vectors"].id.get_chunk_info(i).byte_offset,
                        dtype=fp["vectors"].dtype,  # pyright: ignore[reportAttributeAccessIssue]
                    )
                    for i in range(fp["vectors"].id.get_num_chunks())
                ]
            LOGGER.debug("created memory maps for %s chunks", len(arrays))
            self._mmap_indexer = ChunkIndexer(
                arrays, self._doc_id_to_idx, self._psg_id_to_idx
            )
        return self._mmap_indexer

    def _on_quantizer_set(self) -> None:
        assert self.quantizer is not None

        # serialize the quantizer and store it on disk
        with h5py.File(self._index_file, "a") as fp:
            if "quantizer" in fp:
                del fp["quantizer"]

            meta, attributes, data = self.quantizer.serialize()
            fp.create_group("quantizer/meta").attrs.update(meta)
            fp.create_group("quantizer/attributes").attrs.update(attributes)
            data_group = fp.create_group("quantizer/data")
            for k, v in data.items():
                data_group.create_dataset(k, data=v)

    def _create_ds(self, fp: h5py.File, dim: int, dtype: np.dtype) -> None:
        """Create the HDF5 datasets for vectors and IDs.

        :param fp: Index file (write permissions).
        :param dim: Dimension of the vectors.
        :param dtype: Type of the vectors.
        """
        fp.create_dataset(
            "vectors",
            (self._init_size, dim),
            dtype,
            maxshape=(None, dim),
            chunks=(self._chunk_size, dim),
        )
        fp.create_dataset(
            "doc_ids",
            (self._init_size,),
            f"S{self._max_id_length}",
            maxshape=(None,),
            chunks=True,
        )
        fp.create_dataset(
            "psg_ids",
            (self._init_size,),
            f"S{self._max_id_length}",
            maxshape=(None,),
            chunks=True,
        )

    def _get_num_vectors(self) -> int:
        with h5py.File(self._index_file, "r") as fp:
            return fp.attrs["num_vectors"]  # pyright: ignore[reportReturnType]

    def _get_internal_dim(self) -> int | None:
        with h5py.File(self._index_file, "r") as fp:
            if "vectors" in fp:
                return fp["vectors"].shape[1]  # pyright: ignore[reportAttributeAccessIssue]
        return None

    def to_memory(self, batch_size: int | None = None) -> InMemoryIndex:
        """Load the index entirely into memory.

        :param batch_size: Use batches instead of adding all vectors at once.
        :return: The loaded index.
        """
        index = InMemoryIndex(
            query_encoder=self._query_encoder,
            quantizer=self._quantizer,
            mode=self.mode,
            encoder_batch_size=self._encoder_batch_size,
            init_size=len(self),
        )
        with h5py.File(self._index_file, "r") as fp:
            num_vectors = cast("int", fp.attrs["num_vectors"])

            batch_size = batch_size or num_vectors
            for i_low in range(0, num_vectors, batch_size):
                i_up = min(i_low + batch_size, num_vectors)

                doc_ids = fp["doc_ids"].asstr()[i_low:i_up]  # pyright: ignore[reportAttributeAccessIssue]
                psg_ids = fp["psg_ids"].asstr()[i_low:i_up]  # pyright: ignore[reportAttributeAccessIssue]
                vectors = fp["vectors"][i_low:i_up]  # pyright: ignore[reportIndexIssue]

                # IDs that don't exist will be returned as empty strings here
                doc_ids[doc_ids == ""] = None  # pyright: ignore[reportIndexIssue]
                psg_ids[psg_ids == ""] = None  # pyright: ignore[reportIndexIssue]
                index._add(vectors, doc_ids=doc_ids, psg_ids=psg_ids)  # pyright: ignore[reportArgumentType]
        return index

    def _check_ids(
        self,
        doc_ids: IDSequence,
        psg_ids: IDSequence,
        max_doc_id_len: int,
        max_psg_id_len: int,
    ) -> None:
        """Verify that IDs are valid.

        :param doc_ids: Document IDs to add.
        :param psg_ids: Passage IDs to add.
        :param max_doc_id_len: Maximum length (in characters) of document IDs.
        :param max_psg_id_len: Maximum length (in characters) of passage IDs.
        :raises RuntimeError: When a document ID is longer than allowed.
        :raises RuntimeError: When a passage ID is longer than allowed.
        :raises RuntimeError: When a passage ID already exists.
        """
        for doc_id in doc_ids:
            if doc_id is None:
                continue
            if len(doc_id) > max_doc_id_len:
                raise RuntimeError(
                    f"Document ID {doc_id} is longer than the maximum "
                    f"({max_doc_id_len} characters)."
                )
        for psg_id in psg_ids:
            if psg_id is None:
                continue
            if len(psg_id) > max_psg_id_len:
                raise RuntimeError(
                    f"Passage ID {psg_id} is longer than the maximum "
                    f"({max_psg_id_len} characters)."
                )
            if psg_id in self._psg_id_to_idx:
                raise RuntimeError(f"Passage ID {psg_id} already exists.")

    def _add(
        self,
        vectors: np.ndarray,
        doc_ids: IDSequence,
        psg_ids: IDSequence,
    ) -> None:
        with h5py.File(self._index_file, "a") as fp:
            # if this is the first call to _add, no datasets exist
            if "vectors" not in fp:
                self._create_ds(fp, vectors.shape[-1], vectors.dtype)

            # check all IDs first before adding anything
            self._check_ids(
                doc_ids,
                psg_ids,
                fp["doc_ids"].dtype.itemsize,  # pyright: ignore[reportAttributeAccessIssue]
                fp["psg_ids"].dtype.itemsize,  # pyright: ignore[reportAttributeAccessIssue]
            )

            num_new_vecs = vectors.shape[0]
            capacity = fp["vectors"].shape[0]  # pyright: ignore[reportAttributeAccessIssue]

            # check if we have enough space, resize if necessary
            num_cur_vectors = cast("int", fp.attrs["num_vectors"])
            space_left = capacity - num_cur_vectors
            if num_new_vecs > space_left:
                # resize based on chunk size, i.e., allocate as many chunks as necessary
                # for the new data
                new_size = (
                    int((num_cur_vectors + num_new_vecs) / self._chunk_size + 0.5)
                ) * self._chunk_size
                LOGGER.debug("resizing index from %s to %s", capacity, new_size)
                for ds in ("vectors", "doc_ids", "psg_ids"):
                    fp[ds].resize(new_size, axis=0)  # pyright: ignore[reportAttributeAccessIssue]

                # memory maps need to be recreated after resizing
                self._mmap_indexer = None

            # add new document IDs to index and in-memory mappings
            doc_id_idxs, non_null_doc_ids = [], []
            for i, doc_id in enumerate(doc_ids):
                if doc_id is not None:
                    self._doc_id_to_idx[doc_id].append(num_cur_vectors + i)
                    doc_id_idxs.append(num_cur_vectors + i)
                    non_null_doc_ids.append(doc_id)
            fp["doc_ids"][doc_id_idxs] = non_null_doc_ids  # pyright: ignore[reportIndexIssue]

            # add new passage IDs to index and in-memory mappings
            psg_id_idxs, non_null_psg_ids = [], []
            for i, psg_id in enumerate(psg_ids):
                if psg_id is not None:
                    self._psg_id_to_idx[psg_id] = num_cur_vectors + i
                    psg_id_idxs.append(num_cur_vectors + i)
                    non_null_psg_ids.append(psg_id)
            fp["psg_ids"][psg_id_idxs] = non_null_psg_ids  # pyright: ignore[reportIndexIssue]

            # add new vectors
            fp["vectors"][num_cur_vectors : num_cur_vectors + num_new_vecs] = vectors  # pyright: ignore[reportIndexIssue]
            fp.attrs["num_vectors"] += num_new_vecs  # pyright: ignore[reportOperatorIssue]

    def _get_doc_ids(self) -> set[str]:
        return set(self._doc_id_to_idx.keys())

    def _get_psg_ids(self) -> set[str]:
        return set(self._psg_id_to_idx.keys())

    def _get_vectors(self, ids: "Iterable[str]") -> tuple[np.ndarray, list[str]]:
        if self._memory_mapped:
            return self._get_mmap_indexer()(ids, self.mode)

        # if there are no memory maps, retrieve the vectors using h5py
        indices, ids_ = get_indices(
            ids, self.mode, self._doc_id_to_idx, self._psg_id_to_idx
        )
        if len(ids_) == 0:
            return np.array([]), []

        # h5py requires accessing the dataset with sorted indices
        idx_pairs = zip(indices, ids_)
        h5_indices, out_ids = zip(*sorted(idx_pairs, key=lambda x: x[0]))
        h5_indices = list(h5_indices)

        with h5py.File(self._index_file, "r") as fp:
            # reading all vectors at once slows h5py down significantly, so we read them
            # in batches and concatenate
            vectors = np.concatenate(  # pyright: ignore[reportCallIssue]
                [
                    fp["vectors"][  # pyright: ignore[reportIndexIssue, reportArgumentType]
                        h5_indices[i : i + self._max_indexing_size]
                    ]
                    for i in range(0, len(h5_indices), self._max_indexing_size)
                ]
            )
            return vectors, list(out_ids)

    def _batch_iter(
        self, batch_size: int
    ) -> "Iterator[tuple[np.ndarray, IDSequence, IDSequence]]":
        with h5py.File(self._index_file, "r") as fp:
            num_vectors = cast("int", fp.attrs["num_vectors"])
            for i in range(0, num_vectors, batch_size):
                j = min(i + batch_size, num_vectors)
                doc_ids = fp["doc_ids"].asstr()[i:j]  # pyright: ignore[reportAttributeAccessIssue]
                psg_ids = fp["psg_ids"].asstr()[i:j]  # pyright: ignore[reportAttributeAccessIssue]
                doc_ids[doc_ids == ""] = None  # pyright: ignore[reportIndexIssue]
                psg_ids[psg_ids == ""] = None  # pyright: ignore[reportIndexIssue]
                yield (
                    fp["vectors"][i:j],  # pyright: ignore[reportIndexIssue, reportReturnType]
                    doc_ids.tolist(),  # pyright: ignore[reportAttributeAccessIssue]
                    psg_ids.tolist(),  # pyright: ignore[reportAttributeAccessIssue]
                )

    @classmethod
    def load(
        cls,
        index_file: "Path",
        query_encoder: "Encoder | None" = None,
        mode: Mode = Mode.MAXP,
        encoder_batch_size: int = 32,
        memory_mapped: bool = False,
        max_indexing_size: int = 2**10,
    ) -> "OnDiskIndex":
        """Open an existing index on disk.

        :param index_file: The index file to open.
        :param query_encoder: The query encoder.
        :param mode: The ranking mode.
        :param encoder_batch_size: Batch size for the query encoder.
        :param memory_mapped: Use memory maps for retrieval of vectors if possible.
        :param max_indexing_size:
            Maximum number of vectors to retrieve from the HDF5 dataset at once.
        :return: The index.
        """
        LOGGER.debug("reading file %s", index_file)

        index = cls.__new__(cls)
        super(OnDiskIndex, index).__init__(
            query_encoder=query_encoder,
            quantizer=None,
            mode=mode,
            encoder_batch_size=encoder_batch_size,
        )
        index._index_file = index_file.absolute()
        index._memory_mapped = memory_mapped
        index._max_indexing_size = max_indexing_size

        index._mmap_indexer = None

        # deserialize quantizer if any
        with h5py.File(index_file, "r") as fp:
            if "quantizer" in fp:
                index._quantizer = Quantizer.deserialize(
                    dict(fp["quantizer/meta"].attrs),  # pyright: ignore[reportArgumentType]
                    dict(fp["quantizer/attributes"].attrs),  # pyright: ignore[reportArgumentType]
                    {k: v[:] for k, v in fp["quantizer/data"].items()},  # pyright: ignore[reportAttributeAccessIssue]
                )

            # read ID mappings
            index._doc_id_to_idx = defaultdict(list)
            index._psg_id_to_idx = {}

            num_vectors = cast("int", fp.attrs["num_vectors"])
            if num_vectors == 0:
                return index

            doc_id_iter = fp["doc_ids"].asstr()[:num_vectors]  # pyright: ignore[reportAttributeAccessIssue]
            psg_id_iter = fp["psg_ids"].asstr()[:num_vectors]  # pyright: ignore[reportAttributeAccessIssue]
            for i, (doc_id, psg_id) in tqdm(
                enumerate(zip(doc_id_iter, psg_id_iter)),
                total=num_vectors,
            ):
                if len(doc_id) > 0:
                    index._doc_id_to_idx[doc_id].append(i)
                if len(psg_id) > 0:
                    index._psg_id_to_idx[psg_id] = i
        return index
